from enum import Enum


class CoreEvents(str, Enum):
    BEFORE_APPS_LOAD = 'before_apps_load'
    APPS_LOADED = 'apps_loaded'
    RELOAD_APPS = 'reload_apps'


class CoreEventBus:
    __slots__ = ('events_subs',)

    def __init__(self):
        self.events_subs = {k: set() for k in CoreEvents}

    def on(self, event, cb=None):
        if not cb:
            return lambda cb: self.on(event, cb)
        self.events_subs[CoreEvents(event)].add(cb)
        return cb

    def emit(self, event, *args, **kw):
        [cb(*args, **kw) for cb in self.events_subs[CoreEvents(event)].copy()]

    def off(self, event, cb):
        if event == '*':
            evnts = CoreEvents
        else:
            evnts = [CoreEvents(event)]
        es = self.events_subs
        [es[e].difference_update({cb}) for e in evnts]


core_event_bus = CoreEventBus()



from .template import Template
from .session import Session, GroupSession
from .dal import DAL
from .url import URL
from .xauth import XAuth, AuthGuard, AuthErr, CurrentUser
from .env import Env


__all__ = (
    'Template',
    'Session', 'GroupSession',
    'DAL',
    'URL',
    'XAuth',
    'AuthGuard',
    'AuthErr',
    'CurrentUser',
    'Env',
)



import sys
import os

from . import globs
from .core_events import core_event_bus, CoreEvents
from .install import install_args
from .reloader import Reloader


__all__ = (
    'import_apps',
    'reload_apps',
    'wsgi',
)


def import_apps(pyjsaw_installed):
    core_event_bus.emit(CoreEvents.BEFORE_APPS_LOAD)

    Reloader.import_apps()
    if pyjsaw_installed is not None:
        if 'pyjsaw' in Reloader.modules:
            # we have apps/pyjsaw
            # forget pyjsaw_installed
            Reloader.forget_package('pyjsaw')
        else:
            pyjsaw_installed.websaw_main()

    core_event_bus.emit(CoreEvents.APPS_LOADED)


@core_event_bus.on(CoreEvents.RELOAD_APPS)
def reload_apps(*app_names):
    Reloader.reimport_apps(*app_names)
    core_event_bus.emit(CoreEvents.APPS_LOADED)


def wsgi(**kwargs):
    """Initializes everything, loads apps, returns the wsgi app"""
    install_args(kwargs)
    try:
        import pyjsaw
    except ImportError:
        pyjsaw = None
    import_apps(pyjsaw)
    return globs.app



# Copyrighted (c) by Massimo Di Pierro (2017-2022), Valery Kucherov (2022)

import os
import logging
import datetime
import uuid
import sys
import platform
import traceback
import inspect
import cgitb
import linecache

from pydal import DAL, Field

from .utils import safely
from .globs import request


def get_error_snapshot(depth=5):
    """Return a dict describing a given traceback (based on cgitb.text)."""

    etype, evalue, etb = sys.exc_info()
    if isinstance(etype, type):
        etype = etype.__name__

    data = {}
    data["timestamp"] = datetime.datetime.utcnow().isoformat().replace("T", " ")
    data["python_version"] = sys.version
    platform_keys = [
        "machine",
        "node",
        "platform",
        "processor",
        "python_branch",
        "python_build",
        "python_compiler",
        "python_implementation",
        "python_revision",
        "python_version",
        "python_version_tuple",
        "release",
        "system",
        "uname",
        "version",
    ]

    data["platform_info"] = {key: getattr(platform, key)() for key in platform_keys}
    data["os_environ"] = {key: str(value) for key, value in os.environ.items()}
    data["traceback"] = traceback.format_exc()
    data["exception_type"] = str(etype)
    data["exception_value"] = str(evalue)

    # Loopover the stack frames
    items = inspect.getinnerframes(etb, depth)
    del etb  # Prevent circular references that would cause memory leaks
    data["stackframes"] = stackframes = []

    for frame, file, lnum, func, lines, idx in items:
        file = file and os.path.abspath(file) or "?"
        args, varargs, varkw, locals = inspect.getargvalues(frame)
        # Basic frame information
        f = {"file": file, "func": func, "lnum": lnum}
        f["code"] = lines
        # FIXME: disable this for now until we understand why this goes into infinite loop
        if False:
            line_vars = cgitb.scanvars(
                lambda: linecache.getline(file, lnum), frame, locals
            )
            # Dump local variables (referenced in current line only)
            f["vars"] = {
                key: repr(value)
                for key, value in locals.items()
                if not key.startswith("__")
            }
        stackframes.append(f)

    return data


class SimpleErrorLogger:
    def log(self, app_name, snapshot):
        """logs the error"""
        logging.error("%s error:\n%s" % (app_name, snapshot["traceback"]))
        return None


class ErrorLogger:

    """
    To create your own custom logger for an app:

    class MyLogger:
        def log(app_name, error_snap_shop):
            ...
            return ticket_uuid

    error_logger.plugins['app_name'] = MyLogger()
    """

    def __init__(self):
        self.fallback_logger = SimpleErrorLogger()
        self.database_logger = None
        self.plugins = {}

    def initialize(self):
        """try inizalize database if we have service folder"""
        self.database_logger = safely(DatabaseErrorLogger, log=True)

    def _get_logger(self, app_name):
        """get the appropriate logger for the app"""
        return (
            self.plugins.get(app_name) or self.database_logger or self.fallback_logger
        )

    def log(self, app_name, error_snapshot):
        """log the error snapshot"""
        logger = self._get_logger(app_name)
        ticket_uuid = safely(lambda: logger.log(app_name, error_snapshot))
        if not ticket_uuid:
            self.fallback_logger.log(app_name, error_snapshot)
        return ticket_uuid


class DatabaseErrorLogger:
    def __init__(self):
        """creates the websaw_error table in the service database"""
        uri = os.environ["WEBSAW_SERVICE_DB_URI"]
        folder = os.environ["WEBSAW_SERVICE_FOLDER"]
        self.db = DAL(uri, folder=folder)
        self.db.define_table(
            "websaw_error",
            Field("uuid"),
            Field("app_name"),
            Field("method"),
            Field("path", "string"),
            Field("timestamp", "datetime"),
            Field("client_ip", "string"),
            Field("error", "string"),
            Field("snapshot", "json"),
        )
        self.db.commit()

    def log(self, app_name, error_snapshot):
        """store error snapshot (ticket) in the database"""
        # FIXME:
        # self.db._adapter.reconnect() ?
        ticket_uuid = str(uuid.uuid4())
        try:
            self.db.websaw_error.insert(
                uuid=ticket_uuid,
                app_name=app_name,
                method=request.method,
                path=request.path,
                timestamp=datetime.datetime.utcnow(),
                client_ip=request.environ.get("REMOTE_ADDR"),
                error=error_snapshot["exception_value"],
                snapshot=error_snapshot,
            )
            self.db.commit()
            return ticket_uuid
        except Exception as err:
            logging.error(str(err))
            self.db.rollback()
            return None

    def get(self, ticket_uuid=None):
        """retrieve a ticket from error database"""
        db = self.db
        if ticket_uuid:
            query, orderby = db.websaw_error.uuid == ticket_uuid, None
            rows = db(query).select(orderby=orderby, limitby=(0, 1)).as_list()
        else:
            orderby = ~db.websaw_error.timestamp
            groupby = db.websaw_error.path | db.websaw_error.error
            query = (
                db.websaw_error.timestamp
                > datetime.datetime.now() - datetime.timedelta(days=7)
            )
            fields = [field for field in db.websaw_error if not field.type == "json"]
            fields.append(db.websaw_error.id.count())
            list_rows = (
                db(query).select(*fields, orderby=orderby, groupby=groupby).as_list()
            )
            rows = []
            for item in list_rows:
                row = item["websaw_error"]
                row["count"] = item["_extra"][str(db.websaw_error.id.count())]
                rows.append(row)
        return rows if not ticket_uuid else rows[0] if rows else None

    def clear(self):
        """erase all tickets from database"""
        db = self.db
        db(db.websaw_error).delete()
        self.db.commit()


error_logger = ErrorLogger()




import ombott
from .globs import request, response
from ._core import (
    import_apps,
    reload_apps,
    wsgi,
)

from .utils import (
    redirect,
    Cache,
)

from .reloader import Reloader
from .app import BaseApp
from .context import BaseContext
from .fixture import Fixture

from .render import render

from .exceptions import HTTP, WebsawException

abort = ombott.abort


__all__ = (
    'request',
    'response',
    'abort',

    'Reloader',
    'import_apps',
    'reload_apps',
    'wsgi',
    'BaseContext',
    'BaseApp',
    'Fixture',

    # 'URL',
    'redirect',
    'Cache',

    'render',

    'HTTP',
    'WebsawException',
)


import uuid
import json
import jwt
import time

from ..core.globs import current_config
from ..core import Fixture
from ..core import BaseContext


class Session(Fixture):
    token_env_key = "session_token"

    # All apps share the same default secret if not specified.

    def __init__(
        self,
        secret=None,
        expiration=None,
        algorithm="HS256",
        storage=None,
        same_site="Lax",
    ):
        """
        secret is the shared key used to encrypt the session (using algorithm)
        expiration is in seconds
        (optional) storage must have a get(key) and set(key,value,expiration) methods
        if not provided session is stored in jwt cookie else the jwt is stored in storage
        and its uuid key is stored in the cookie
        """
        self.secret = secret  # see app_mounted
        self.expiration = int(expiration) if expiration is not None else None
        self.algorithm = algorithm
        self.storage = storage
        self.same_site = same_site

    def app_mounted(self, ctx):
        if not self.secret:
            self.secret = current_config.session_secret
        assert self.secret

    def take_on(self, ctx: BaseContext):
        self.data.ctx = ctx
        self.load(ctx)

    def take_off(self, ctx):
        if self.data.changed:
            self.save()

    @staticmethod
    def get_session_cookie_name(app_data):
        return f'{app_data.app_name}_session'

    def initialize(self, request, response, session_cookie_name, data=None, secure=False, storage=None):
        local = self.data
        local.request = request
        local.response = response
        local.changed = False
        local.data = data or {}
        local.session_cookie_name = session_cookie_name  # f"{app_name}_session"
        local.secure = secure
        local.storage = storage

    def load(self, ctx: BaseContext):
        request = ctx.request
        storage = None
        data_in_cookie = True
        if self.storage:
            storage = ctx.ask(self.storage)
            if storage is None:
                raise KeyError(f'Storage is not provided by context: {self.storage}')
            data_in_cookie = False

        session_cookie_name = self.get_session_cookie_name(ctx.app_data)

        env_get = request.environ.get
        secure = "https" == (
            env_get('HTTP_X_FORWARDED_PROTO') or env_get('wsgi.url_scheme')
        )

        self.initialize(
            request=request,
            response=ctx.response,
            session_cookie_name=session_cookie_name,
            secure=secure,
            storage=storage
        )

        token_data = (
            request.get_cookie(session_cookie_name)
            or ctx.env.get(self.token_env_key)
        )

        # fast exit
        if not token_data:
            return

        if isinstance(token_data, str):
            token_data = token_data.encode()

        data = None
        if storage is not None:
            json_data = storage.get(token_data)
            if json_data:
                try:
                    data = json.loads(json_data)
                except json.JSONDecodeError:
                    pass
        else:
            try:
                data = jwt.decode(
                    token_data, self.secret, algorithms=[self.algorithm]
                )
            except jwt.PyJWTError:
                pass

        if data is None:
            return

        is_valid = True

        if session_cookie_name != data.get("session_cookie_name"):
            is_valid = False
        if is_valid and "uuid" not in data:
            is_valid = False
        if is_valid and data_in_cookie and self.expiration is not None:
            is_valid = self.expiration > time.time() - data.get("timestamp", 0)
        if is_valid and secure is not data.get("secure"):
            is_valid = False

        if not is_valid:
            self.clear()
        else:
            self.data.data = data

    def get_data(self):
        return self.data.data

    def save(self):
        local = self.data
        data = local.data
        response = local.response
        data["timestamp"] = time.time()
        data["session_cookie_name"] = local.session_cookie_name
        data["secure"] = local.secure
        if "uuid" not in data:
            data["uuid"] = str(uuid.uuid1())
        if local.storage is not None:
            cookie_data = data["uuid"]
            local.storage.set(cookie_data, json.dumps(data), self.expiration)
        else:
            cookie_data = jwt.encode(
                data, self.secret, algorithm=self.algorithm
            )
            if isinstance(cookie_data, bytes):
                cookie_data = cookie_data.decode()

        response.set_cookie(
            local.session_cookie_name,
            cookie_data,
            path="/",
            secure=local.secure,
            samesite=self.same_site,
        )

    def get(self, key, default=None):
        return self.data.data.get(key, default)

    def __getitem__(self, key):
        return self.data.data[key]

    def __delitem__(self, key):
        data = self.data.data
        if key in data:
            self.data.changed = True
            del data[key]

    def __setitem__(self, key, value):
        self.data.changed = True
        self.data.data[key] = value

    def __contains__(self, k):
        return k in self.keys()

    def keys(self):
        return self.data.data.keys()

    def __iter__(self):
        yield from self.data.data.items()

    def clear(self):
        """Produces a brand-new session."""
        local = self.data
        local.changed = True
        data = local.data
        data.clear()
        data["uuid"] = str(uuid.uuid1())
        data["secure"] = local.secure


class GroupSession(Session):

    @staticmethod
    def get_session_cookie_name(app_data):
        return f'{app_data.group_name}_session'



import functools
from types import SimpleNamespace
from typing import List

from . import globs
from .context import BaseContext
from .exceptions import FixtureProcessError
from .reloader import Reloader
from .static_registry import static_registry


def _dummy_exception_handler(ctx: BaseContext, exc: Exception):
    raise exc


class Fixtured:
    def __init__(self, h, fixt: List[str]):
        if isinstance(h, self.__class__):
            fixt = [*fixt, *h.fixt]
            h = h.h
        self.h = h
        self.fixt = fixt
        functools.update_wrapper(self, h)

    def __call__(self, *a, **kw):
        return self.h(*a, **kw)


class BaseApp:

    static_registry = static_registry
    add_route = staticmethod(globs.app.add_route)
    reloader = Reloader

    def __init__(
        self,
        default_config,
        default_ctx: BaseContext,
    ):
        self.default_config = default_config
        self.default_ctx = default_ctx
        self._registered = {}
        self._mixins: List['BaseApp'] = []

    def mixin(self, *mixins):
        self._mixins.extend(mixins)
        self.default_ctx.extend(*[m.default_ctx for m in mixins])

    def _register(self, fun, route_args, fixtures=None):
        meta = self._registered.setdefault(
            fun, SimpleNamespace(routes_args=[], fixtures=[])
        )
        meta.routes_args.append(route_args)
        if fixtures is not None:
            meta.fixtures.extend(fixtures)

    def route(self, rule, method='GET', name=None, **kw):
        args = (rule, method, name)

        def decorator(h):
            fixt = None
            if isinstance(h, Fixtured):
                fixt = h.fixt
                h = h.h
            self._register(h, (args, kw), fixt)
            return h
        return decorator

    def use(self, *fixt):
        fixt = [self.default_ctx.get_or_make_fixture_key(f) for f in fixt]

        def decorator(h):
            if isinstance(h, Fixtured):
                h.fixt[:] = [*fixt, *h.fixt]
                return h
            return Fixtured(h, fixt)
        return decorator

    def mount(
        self,
        config: dict = None,
        context: BaseContext = None,
        render_map: dict = None,
        exception_handler=None
    ):
        if context is None:
            context = self.default_ctx
        if config is None:
            config = self.default_config

        render_map: dict = config.get('render_map')
        exception_handler = config.get('exception_handler')

        context = context.clone()
        app_data = context.app_data = SimpleNamespace(
            routes=[],
            named_routes={},
            **config
        )
        for raw_h, meta, mixin_data in self._iter_registered():
            h = self.make_handler(raw_h, meta.fixtures, context, render_map, exception_handler, mixin_data)
            for route_args, route_kw in meta.routes_args:
                self._mount_route(context.app_data, h, route_args, route_kw)

        # mount app static
        static_rule, static_h = self.static_registry.make_rule_and_handler(
            f'{app_data.static_base_url}/static', app_data.static_folder, app_data.app_name
        )
        if static_rule is not None:
            self._mount_route(context.app_data, static_h, (static_rule, 'GET', None), {})

        # mount mixins static as /{app_name}/static/{mixin_name}/
        for m in self._mixins:
            m_cfg = m.default_config
            m_name = m_cfg['app_name']
            static_base_url = f'{app_data.base_url}/static/mxn/{m_name}'
            static_rule, static_h = self.static_registry.make_rule_and_handler(
                static_base_url, m_cfg['static_folder'], app_data.app_name
            )
            if static_rule is not None:
                self._mount_route(context.app_data, static_h, (static_rule, 'GET', None), {})

        # register
        self.reloader.register_app_data(context.app_data)
        context.app_mounted()
        return context

    def _iter_registered(self):
        for m in reversed(self._mixins):
            for raw_h, meta in m._registered.items():
                yield raw_h, meta, SimpleNamespace(**m.default_config)

        for raw_h, meta in self._registered.items():
            yield raw_h, meta, None

    @staticmethod
    def make_handler(h, fixtures, ctx: BaseContext, render_map: dict = None, exception_handler=None, mixin_data=None):

        hooks = False
        if fixtures:
            hooks = {
                fkey: fobj for fkey, fobj
                in ([fkey, getattr(ctx, fkey)] for fkey in fixtures)
                if fobj.is_hook
            } or False
        else:
            fixtures = False

        if exception_handler is None:
            exception_handler = _dummy_exception_handler

        @functools.wraps(h)
        def handler(**kw):
            exc = None
            ctx.initialize()
            ctx.mixin_data = mixin_data
            try:
                if fixtures:
                    ctx.use_fixtures(fixtures, hooks)
                ctx.output = h(ctx, **kw)
            except FixtureProcessError:
                pass
            except Exception as exc_:
                exc = exc_
            ctx.finalize(exc)
            if ctx.exception is not None:
                exception_handler(ctx, ctx.exception)

            if render_map:
                output = ctx.output
                render = render_map.get(type(output), False)
                if render:
                    ctx.output = render(ctx, output)
            return ctx.output
        return handler

    @staticmethod
    def _get_abs_url(base_url, path):
        if not path:
            return base_url
        if path[0] != '/':
            path = f'{base_url}/{path}'
        return path

    def _mount_route(self, app_data, fun, route_args, route_kw):
        path, method, name = route_args

        is_index = path == 'index'
        path = self._get_abs_url(app_data.base_url, path)
        route = self.add_route(path, method, fun, **route_kw)
        app_data.routes.append(route)
        if name:
            if name in app_data.named_routes:
                raise KeyError(f'The route name already in use: {name}')
            app_data.named_routes[name] = route

        if is_index:
            route = self.add_route(
                path[:-len('/index')] or '/', method, fun, **route_kw
            )
            app_data.routes.append(route)



import copy
from pydal import DAL as _DAL
from threadsafevariable import ThreadSafeVariable

from ..core.core_events import core_event_bus, CoreEvents
from ..core.fixture import Fixture


# this global object will be used to store their state to restore it for every http request
ICECUBE = {}


@core_event_bus.on(CoreEvents.APPS_LOADED)
def update_icecube(*a, **kw):
    ICECUBE.update(ThreadSafeVariable.freeze())


class DAL(_DAL, Fixture):

    reconnect_on_request = True

    def take_on(self, ctx):
        if self.reconnect_on_request:
            self._adapter.reconnect()
        ThreadSafeVariable.restore(ICECUBE)

    def take_off(self, ctx):
        if ctx.exception:
            self.rollback()
        else:
            self.commit()
        self._adapter.close(action=None, really=False)


# make sure some variables in pydal are thread safe
def thread_safe_pydal_patch():
    Field = _DAL.Field
    tsafe_attrs = [
        "readable",
        "writable",
        "default",
        "update",
        "requires",
        "widget",
        "represent",
    ]
    [setattr(Field, a, ThreadSafeVariable()) for a in tsafe_attrs]

    # hack 'copy.copy' behavior, since it makes a shallow copy,
    # but ThreadSafe-attributes (see above) are class-level, so:
    # no copy -> no attr in ICECUBE for the fresh one -> gevent-error on try to access to any of ThreadSafe-attributes
    def field_copy(self):
        # to prevent infinite recursion
        # temporarily set __copy__ to None
        me = self.__class__.__copy__
        self.__class__.__copy__ = None
        clone = copy.copy(self)
        self.__class__.__copy__ = me
        [setattr(clone, a, getattr(self, a)) for a in tsafe_attrs]
        return clone

    # to avoid possible future problems
    if hasattr(Field, "__copy__"):
        raise RuntimeError("code fix required!")
    setattr(Field, "__copy__", field_copy)


thread_safe_pydal_patch()



import urllib
import sys
import os
import logging
import importlib
import threading
import functools
import time
import numbers
import datetime
import enum
import types
import json

from .globs import response
from .exceptions import HTTP

url_quote = urllib.parse.quote


def redirect(location):
    """our redirect does not delete cookies and headers like bottle.HTTPResponse does;
    it is considered a success, not failure"""
    response.headers["Location"] = str(location)
    raise HTTP(303)


class MetaPathRouter:
    """
    Instances of this class makes alias for a package name,
    in other words instruct the import system to route request
    for a package alias, i.e.:

        MetaPathRouter("pkg", "pkg_alias")
        import pkg_alias.sub

    works as

        import pkg.sub

    author: Paolo Pastori
    """

    def __init__(self, pkg, pkg_alias="apps"):
        assert pkg_alias
        assert pkg
        if pkg != pkg_alias:
            self.pkg_alias = pkg_alias
            self.pkg = pkg
            # register as path finder
            sys.meta_path.append(self)

    def find_spec(self, fullname, path=None, target=None):
        if fullname == self.pkg_alias and path is None:
            spec = importlib.util.find_spec(self.pkg)
            if spec:
                spec.name = fullname
                spec.loader = importlib.machinery.SourceFileLoader(
                    fullname, spec.origin
                )
                return spec


def safely(func, exceptions=(Exception,), log=False, default=None):
    """
    runs the funnction and returns True on success,
    False if one of the exceptions is raised
    """
    try:
        return func()
    except exceptions as err:
        if log:
            logging.warn(str(err))
        return default() if callable(default) else default


def module2filename(module):
    filename = os.path.join(*module.split(".")[1:])
    filename = (
        os.path.join(filename, "__init__.py")
        if not filename.count(os.sep)
        else filename + ".py"
    )
    return filename


########################################################################################
# Implement a O(1) LRU cache and memoize with expiration and monitoring (using linked list)
#########################################################################################


class Node:
    def __init__(self, key=None, value=None, t=None, m=None, prev=None, next=None):
        self.key, self.value, self.t, self.m, self.prev, self.next = (
            key,
            value,
            t,
            m,
            prev,
            next,
        )


class Cache:
    """
    O(1) caching object that remembers the 'size' most recent values
    Example:

        cache = Cache(size=1000)
        h = cache.get(filename, lambda: hash(
            open(filename).read()), 60, lambda: os.path.getmtime())

    (computes and cashes the hash of file filename but only reads the file if mtime changes and
     does not check the mtime more oftern than every 60. caches the 1000 most recent hashes)
    """

    def __init__(self, size=1000):
        self.free = size
        self.head = Node()
        self.tail = Node()
        self.head.next = self.tail
        self.tail.prev = self.head
        self.mapping = {}
        self.lock = threading.Lock()

    def get(self, key, callback, expiration=3600, monitor=None):
        """If key not stored or key has expired and monitor == None or monitor() value has changed, returns value = callback()"""
        node, t0 = self.mapping.get(key), time.time()
        with self.lock:
            if node:
                # if a node was found remove it from storage
                value, t, node.next.prev, node.prev.next = (
                    node.value,
                    node.t,
                    node.prev,
                    node.next,
                )
            else:
                self.free -= 1
        # check if something may invalidate cache
        m = monitor() if monitor else None
        # check if cache expired
        if node and node.t + expiration < t0:
            # if cache should always be invalidated or m changed
            if m is None or node.m != m:
                # ignore the value found
                node = None
        if node is None:
            value, t = callback(), t0
        # add the new node back into storage
        with self.lock:
            new_node = Node(key, value, t, m, prev=self.head, next=self.head.next)
            self.mapping[key] = self.head.next = new_node.next.prev = new_node
            if self.free < 0:
                last_node = self.tail.prev
                self.tail.prev, last_node.prev.next = last_node.prev, self.tail
                del self.mapping[last_node.key]
                self.free += 1
        return value

    def memoize(self, expiration=3600):
        def decorator(func):
            @functools.wraps(func)
            def memoized_func(*args, **kwargs):
                key = "%s:%s:%s:%s" % (func.__module__, func.__name__, args, kwargs)
                return self.get(
                    key,
                    lambda args=args, kwargs=kwargs: func(*args, **kwargs),
                    expiration=expiration,
                )

            return memoized_func

        return decorator


#########################################################################################
# A Better JSON Serializer
#########################################################################################


def objectify(obj):
    """converts the obj(ect) into a json serializable object"""
    if isinstance(obj, numbers.Integral):
        return int(obj)
    elif isinstance(obj, (numbers.Rational, numbers.Real)):
        return float(obj)
    elif isinstance(obj, (datetime.date, datetime.datetime, datetime.time)):
        return obj.isoformat().replace("T", " ")
    elif isinstance(obj, (str, dict)):
        return obj
    elif isinstance(obj, enum.Enum):  # Enum class handled specially to address self reference in __dict__
        return dict(name=obj.name, value=obj.value, __class__=obj.__class__.__name__)
    else:
        jsonable_meth = (
            getattr(obj, "as_list", None) or  # noqa
            getattr(obj, "as_dict", None) or  # noqa
            getattr(obj, "xml", None)
        )
        if jsonable_meth:
            return jsonable_meth()
        elif hasattr(obj, "__iter__") or isinstance(obj, types.GeneratorType):
            return list(obj)
        elif hasattr(obj, "__dict__") and hasattr(obj, "__class__"):
            d = dict(obj.__dict__)
            d["__class__"] = obj.__class__.__name__
            return d
    return str(obj)


def dumps(obj, sort_keys=True, indent=2, ensure_ascii=False):
    return json.dumps(obj, default=objectify, sort_keys=sort_keys, indent=indent, ensure_ascii=ensure_ascii)





import http.client
import json

from .core import globs
from .core.render import render


ERROR_PAGES = {
    "*": (
        '<html><head><style>body{color:white;text-align: center;background-color:[[=color]];font-family:serif} '
        'h1{font-size:6em;margin:16vh 0 8vh 0} h2{font-size:2em;margin:8vh 0} '
        'a{color:white;text-decoration:none;font-weight:bold;padding:10px 10px;border-radius:10px;border:2px solid #fff;transition: all .5s ease} '
        'a:hover{background:rgba(0,0,0,0.1);padding:10px 30px}</style></head>'
        '<body><h1>[[=code]]</h1><h2>[[=message]]</h2>[[if button_text:]]<a href="[[=href]]">[[=button_text]]</a>[[pass]]</body></html>'
    ),
}


def error_page(code, button_text=None, href="#", color=None, message=None):
    request = globs.request
    message = http.client.responses[code].upper() if message is None else message
    color = (
        {"4": "#F44336", "5": "#607D8B"}.get(str(code)[0], "#2196F3")
        if not color
        else color
    )
    context = dict(
        code=code, message=message, button_text=button_text, href=href, color=color
    )
    # if client accepts 'application/json' - return json
    if request.is_json_requested:
        globs.response.status = code
        return json.dumps(context)
    # else - return html error-page
    content = ERROR_PAGES.get(code) or ERROR_PAGES["*"]
    return render(content=content, context=context, delimiters="[[ ]]")



import threading
from types import SimpleNamespace
from typing import Optional, Any


class Fixture:

    # Central thread local storage to hold local data of all fixtures
    _local = threading.local()

    # Should the fixture be treated as a set of hooks - take_on/take_off.
    # If it is set to True, then 'take_off' method is called
    # regardless of whether 'take_on' method was called
    is_hook = False

    # The name of the fixture in the Context.
    # It is intended to be used in generic fixtures (e.g. Tempalte)
    # and supposed to be set in __init__
    context_key: Optional[str]

    @classmethod
    def initialize_safe_storage(cls):
        """Initialize the central thread local storage of all fixtures.

        We do it with one shot!
        """
        cls._local.fixtures_data = {}

    @classmethod
    def prepare_for_use(cls, fixture: 'Fixture'):
        """Initialize the thread local storage for concrete 'fixture'."""
        cls._local.fixtures_data[fixture] = SimpleNamespace()

    @property
    def data(self) -> SimpleNamespace:
        """Return the thread local fixture storage."""
        return self._local.fixtures_data[self]

    def app_mounted(self, ctx):
        """Is called when app is mounted."""
        ...

    def take_on(self, ctx) -> Optional[Any]:
        """Is called when the fixture is accessed as the context attribute.

        This hook is intended for the fixture initialization when it is accessed
        the first time during request processing.
        The return value is what the consumer will get when accessing the fixture
        as the context attribute. So it should return something useful
        e.g. opened file-descriptor or db-connection, if it returns None
        then the fixture itself will be used.
        The return value is cached in the context until take_off-hook is called.

        Note: This hook is only called if it's the request processing:

        @app.route('index')
        @app.use(ctxd.foo)  # won't be called as it's app loading stage
        def some(ctx):
            ctx.foo         # will be called as it's request processing

        """
        ...

    def take_off(self, ctx):
        """Is called at the end of request processing.

        if self.is_hook is set to False then it is only called if take_on-hook was called

        This hook is for cleanup/tear down action (e.g. to close a file or db-connection)
        """
        ...
